use petgraph::{prelude::*, algo::dijkstra::*, visit::Reversed};

fn char_elevation(c: char) -> i16 {
    let elevation = match c {
        'S' => 'a',
        'E' => 'z',
        c   => c,
    };

    ((elevation as u8) - b'a') as i16
}

fn travelable(dest: char, src: char) -> bool {
    char_elevation(src) + 1 >= char_elevation(dest)
}

fn main() {
    const INPUT: &str = include_str!("../input.txt");
    
    let grid = INPUT.lines().collect::<Vec<&str>>();
    let grid = grid.iter().map(|s| { s.chars().collect::<Vec<char>>() }).collect::<Vec<Vec<char>>>();

    let mut graph: DiGraphMap<(usize, usize), u8> = DiGraphMap::new();

    let height = grid.len();
    let width = grid[0].len();

    let mut start: (usize, usize) = (0, 0);
    let mut end: (usize, usize) = (0, 0);
    let mut starts: Vec<(usize, usize)> = Vec::new();

    for i in 0..height {
        for j in 0..width {
            graph.add_node((i, j));

            let c = grid[i][j];
            if c == 'S' {
                start = (i, j);
                starts.push((i, j));
            } else if c == 'E' {
                end = (i, j);
            } else if c == 'a' {
                starts.push((i, j));
            }
        }
    }

    for i in 0..height {
        for j in 0..width {
            let c = grid[i][j];

            let up = i.saturating_sub(1);
            let down = i + 1;
            let left = j.saturating_sub(1);
            let right = j + 1;

            if left != j && travelable(grid[i][left], c) {
                graph.add_edge((i, j), (i, left), 1);
            }
            if right < width && travelable(grid[i][right], c) {
                graph.add_edge((i, j), (i, right), 1);
            }
            if up != i && travelable(grid[up][j], c) {
                graph.add_edge((i, j), (up, j), 1);
            }
            if down < height && travelable(grid[down][j], c) {
                graph.add_edge((i, j), (down, j), 1);
            }
        }
    }

    let res = dijkstra(&graph, start, Some(end), |_| 1);
    println!("{} steps from ({}, {})", res[&end], start.0, start.1);

    let res = dijkstra(Reversed(&graph), end, None, |_| 1);
    let mut min: (i32, usize, usize) = (i32::MAX, 0, 0);
    for start in starts {
        if res.contains_key(&start) && res[&start] < min.0 {
            min = (res[&start], start.0, start.1);
        }
    }
    println!("{} steps from ({}, {})", min.0, min.1, min.2);
}
